package midp.shopper;

import javax.microedition.lcdui.*;
import java.util.Vector;

public class ProductList extends List implements CommandListener 
{
    /**
     * This is the parent component to return to when back is pressed.
     */
    Displayable parent;
    Category    category;
    
    /**
     * The constructor is a simple replication of the base
     * list constructor.
     */
	
    public ProductList(Category category, Displayable parent)
    {
        super(category.name, ChoiceGroup.MULTIPLE);
        this.category = category;
        this.parent = parent;
        int n = category.products.size();
        boolean[] selected = new boolean[n];
        Shopper shopper = Shopper.getInstance();
        for (int i = 0; i < n; i++) { 
            Product product = (Product)category.products.elementAt(i);
            selected[i] = shopper.isSelected(product);
            append(product.name, null);
        }
        setSelectedFlags(selected);
        setCommandListener(this);
        addCommand(Shopper.ADD_CMD);
        addCommand(Shopper.REMOVE_CMD);
        addCommand(Shopper.BACK_CMD);
        Display.getDisplay(shopper).setCurrent(this);
    }

    
    /**
     * Indicates that a command event has occurred on Screen s.
     */
    public void commandAction(Command c, Displayable d)  
    {
        if (c == Shopper.BACK_CMD) {
            int n = size();
            boolean[] selected = new boolean[n];
            getSelectedFlags(selected);
            Shopper shopper = Shopper.getInstance();
            boolean modified = false;
            for (int i = 0; i < n; i++) { 
                Product product = (Product)category.products.elementAt(i);
                modified |= shopper.setSelected(product, selected[i]);
            }    
            if (modified) { 
                shopper.saveSelected();
            }
            Display.getDisplay(shopper).setCurrent(parent);
        } 
        else if (c == Shopper.REMOVE_CMD) {
            int n = size();
            boolean[] selected = new boolean[n];
            getSelectedFlags(selected);
            Shopper shopper = Shopper.getInstance();
            for (int i = 0, j = 0; i < n; i++) { 
                if (selected[i]) { 
                    Product product = (Product)category.products.elementAt(j);
                    shopper.removeProduct(j, product);
                    delete(j);
                } else { 
                    j += 1;
                }
            }
        } 
        else if (c == Shopper.ADD_CMD) {
            new NewProductBox(this);
        }                   
    }
}







