package midp.shopper;

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;

import javax.microedition.rms.*;
import java.util.Vector;
import java.util.Hashtable;

public class Shopper extends MIDlet implements CommandListener  { 
    private static Shopper instance = null;

    private Vector      categories;
    private RecordStore store;

    private static Image[]   categoryImages;
    private static Image     productsImage;
    private static Hashtable imageHash;
    
    byte[] selected;
    byte[] origSelected;

    byte[] predefined;
    int    nPredefined;

    static Command BACK_CMD;
    static Command ADD_CMD;
    static Command SELECT_CMD;
    static Command REMOVE_CMD;
    static Command RENAME_CMD;
    static Command ENTER_CMD;
    static Command QUIT_CMD;

    static final int PREDEFINED_ID = 1;
    static final int SELECTED_ID  = 2;

    private Class[] actions = {
        getClassObject("midp.shopper.NeedList"),
        getClassObject("midp.shopper.CategoryList"),     
        getClassObject("midp.shopper.ResetList")
    };

    public Vector getCategories() { 
        return categories;
    }

    /**
     * This method dynamically loads a class to work around a limitation
     * in the Forte compiler that prevents us from compiling ClassName.class
     * syntax.
     */
    private static Class getClassObject(String className) {
        try {
            return(Class.forName(className));
        }
        catch(Throwable err) {
            System.out.println(err.getClass().getName() + ": " +
                               err.getMessage());
            err.printStackTrace();
        }
        return(null);
    }

    private List actionList;

    static Shopper getInstance()
    {
        return(instance);
    }


    public Shopper() { 
        String locale = getAppProperty("Shopper.Locale");
        if (locale == null || locale.length() == 0) { 
            locale = System.getProperty("microedition.locale");
        }        
        new Locale(locale);
        BACK_CMD   = new Command(Locale.current.back,  Command.BACK, 1);
        ADD_CMD    = new Command(Locale.current.add,   Command.SCREEN, 1);
        SELECT_CMD = new Command(Locale.current.select, Command.ITEM, 1);
        REMOVE_CMD = new Command(Locale.current.remove,Command.ITEM, 1);
        RENAME_CMD = new Command(Locale.current.rename,Command.ITEM, 1);
        ENTER_CMD  = new Command(Locale.current.enter, Command.SCREEN, 1);
        QUIT_CMD  = new Command(Locale.current.quit, Command.BACK, 1);
	categoryImages = new Image[Locale.current.categoryIcons.length];
        imageHash = new Hashtable();
	try 
	{
            productsImage = Image.createImage("/midp/shopper/products.png");
	    for (int i = 0; i < Locale.current.categoryIcons.length; i++) 
	    { 
		categoryImages[i] = Image.createImage("/midp/shopper/" + Locale.current.categoryIcons[i] + ".png");
                imageHash.put(Locale.current.categories[i], categoryImages[i]);
	    }
	} 
	catch (java.io.IOException err) 
	{
	    // ignore the image loading failure the application can recover.
	}
        actionList = new List(Locale.current.name,
                                       List.IMPLICIT,
                                       Locale.current.actionNames,
                                       null);
        actionList.addCommand(SELECT_CMD);
        actionList.addCommand(QUIT_CMD);
        actionList.setCommandListener(this) ;
        instance = this;
    }

    static final String STORE_NAME = "Shopper107";

    private void openStore() {
        try { 
            try { 
                store = RecordStore.openRecordStore(STORE_NAME, false);
                loadData();
            } catch (RecordStoreNotFoundException x) {             
                store = RecordStore.openRecordStore(STORE_NAME, true);
                initializeStore();
            }            
            origSelected = new byte[selected.length];
            System.arraycopy(selected, 0, origSelected, 0, selected.length);
        } catch(Exception x) { 
            criticalException(x);
        }
    }

    private static void pack(byte[] arr, int x) { 
        arr[0] = (byte)(x >> 24);
        arr[1] = (byte)(x >> 16);
        arr[2] = (byte)(x >> 8);
        arr[3] = (byte)x;
    }

    private static int unpack(byte[] arr) { 
        return (arr[0] << 24) | ((arr[1] & 0xFF) << 16) | ((arr[2] & 0xFF) << 8) | (arr[3] & 0xFF);
    }

    private static byte[] str2Bytes(String str) { 
        char[] chars = str.toCharArray();
        byte[] bytes = new byte[chars.length*2];
        for (int i = 0; i < chars.length; i++) { 
            bytes[i*2] = (byte)(chars[i] >> 8);
            bytes[i*2+1] = (byte)chars[i];
        }
        return bytes;
    }

    private void initializeStore() {
        try { 
            categories = new Vector();
            int n = 0;
            for (int i = 0; i < Locale.current.categories.length; i++) {   
                n += 1;
                Category category = new Category(-n, Locale.current.categories[i], categoryImages[i]);
                categories.addElement(category);

                for (int j = 0; j < Locale.current.products[i].length; j++) { 
                    n += 1;
                    Product product = new Product(-n, category, Locale.current.products[i][j]);                
                    category.products.addElement(product);
                }
            }
            nPredefined = n+1;
            predefined = new byte[(n + 8) >> 3];
            store.addRecord(predefined, 0, predefined.length);
            selected = new byte[predefined.length];
            store.addRecord(selected, 0, selected.length);
        } catch(Exception x) { 
            criticalException(x);
        }
    }
        
    private void loadData() throws Exception { 
        Hashtable hash = new Hashtable();
        try { 
            predefined = store.getRecord(PREDEFINED_ID);
            selected = store.getRecord(SELECTED_ID);
            byte[] deleted = predefined;
            while (true) { 
                int n = 0;
                categories = new Vector();
                for (int i = 0; i < Locale.current.categories.length; i++) {   
                    Category category = null;
                    n += 1;
                    if ((n>>3) >= deleted.length || (deleted[n>>3] & (1 << (n & 7))) == 0) { 
                        category = new Category(-n, Locale.current.categories[i], categoryImages[i]);
                        hash.put(new Integer(-n), category);
                        categories.addElement(category);
                    }
                    for (int j = 0; j < Locale.current.products[i].length; j++) { 
                        n += 1;
                        if (category != null && ((n>>3) >= deleted.length || (deleted[n>>3] & (1 << (n & 7))) == 0)) { 
                            Product product = new Product(-n, category, Locale.current.products[i][j]);                
                            category.products.addElement(product);
                        }
                    }
                }
                nPredefined = n+1;
                int delta = ((n + 8) >> 3) - deleted.length;
                if (delta > 0) { 
                    predefined = deleted = new byte[(n + 8) >> 3];
                    store.setRecord(PREDEFINED_ID, deleted, 0, deleted.length);
                    selected = new byte[selected.length + delta + 1];
                    store.setRecord(SELECTED_ID, selected, 0, selected.length);
                    continue;
                }
                break;
            }
            int last = store.getNextRecordID();
            for (int recordId = SELECTED_ID; ++recordId < last;) { 
                byte[] record = null;
                try { 
                    record = store.getRecord(recordId);                    
                } catch (InvalidRecordIDException x) {
                    // record was deleted 
                }
                if (record == null) { 
                    continue;
                }
                int categoryId = unpack(record);
                char[] chars = new char[(record.length - 4) >> 1];
                for (int i = 0; i < chars.length; i++) { 
                    chars[i] = (char)((record[4 + i*2] << 8) | (record[5 + i*2] & 0xFF));
                }                
                String name = new String(chars);
                if (categoryId == 0) { 
                    Category category = new Category(recordId, name, productsImage);
                    categories.addElement(category);
                    hash.put(new Integer(recordId), category);
                } else { 
                    Category category = (Category)hash.get(new Integer(categoryId));
                    if (category == null) { 
                        criticalError("Failed to locate category");
                        continue;
                    }
                    Product product = new Product(recordId, category, name);
                    category.products.addElement(product);
                }
            }
        } catch (Exception x) {
            try { 
                store.closeRecordStore();
                RecordStore.deleteRecordStore(STORE_NAME);
            } catch (Exception xx) {}
            criticalException(x);
        }
    }

    public void storeCategory(Category category) {
        try { 
            byte[] name = str2Bytes(category.name);
            byte[] record = new byte[name.length + 4];
            System.arraycopy(name, 0, record, 4, name.length);
            store.setRecord(category.recordId, record, 0, record.length);
        } catch (Exception x) {
            criticalException(x);
        }
    }

    public Category addCategory(String name) { 
        byte[] nameBytes = str2Bytes(name);
        byte[] record = new byte[nameBytes.length + 4];
        int recordId = 0;
        System.arraycopy(nameBytes, 0, record, 4, nameBytes.length);
        try { 
            recordId = store.addRecord(record, 0, record.length);
        } catch (Exception x) { 
            criticalException(x);
        }
        Category category = new Category(recordId, name, productsImage);
        categories.addElement(category);
        return category;
    }

    public boolean removeCategory(int i) { 
        Category category = (Category)categories.elementAt(i);
        if (category.products.size() != 0) { 
            error(Locale.current.notEmpty);
            return false;
        } else { 
            if (category.isPredefined()) { 
                int j = -category.recordId;
                predefined[j>>3] |= (byte)(1 << (j & 7));
                try { 
                    store.setRecord(PREDEFINED_ID, predefined, 0, predefined.length);            
                } catch (Exception x) { 
                    criticalException(x);
                }
            } else { 
                try { 
                    store.deleteRecord(category.recordId);            
                } catch (Exception x) { 
                    criticalException(x);
                }
            }
            categories.removeElementAt(i);
            return true;
        }
    }

    public Product addProduct(String name, Category category) { 
        byte[] nameBytes = str2Bytes(name);
        byte[] record = new byte[nameBytes.length + 4];
        int recordId = 0;
        System.arraycopy(nameBytes, 0, record, 4, nameBytes.length);
        pack(record, category.recordId);        
        try { 
            recordId = store.addRecord(record, 0, record.length);
        } catch (Exception x) { 
            criticalException(x);
        }
        Product product = new Product(recordId, category, name);
        category.products.addElement(product);
        int i = (nPredefined + recordId + 7) >> 3;
        while (i >= selected.length) { 
            byte[] newSelected = new byte[selected.length*2];
            System.arraycopy(selected, 0, newSelected, 0, selected.length);
            selected = newSelected;
            try { 
                store.setRecord(SELECTED_ID, newSelected, 0, newSelected.length);
            } catch (Exception x) { 
                criticalException(x);
            }
        }
        return product;
    }

    public void removeProduct(int i, Product product) { 
        if (product.isPredefined()) { 
            int j = -product.recordId;
            predefined[j>>3] |= (byte)(1 << (j & 7));
            try { 
                store.setRecord(PREDEFINED_ID, predefined, 0, predefined.length);            
            } catch (Exception x) { 
                criticalException(x);
            }
        } else { 
            try { 
                store.deleteRecord(product.recordId);            
            } catch (Exception x) { 
                criticalException(x);
                return;
            }
        }
        product.category.products.removeElementAt(i);
    }
        
    public void restoreData() { 
        System.arraycopy(origSelected, 0, selected, 0, origSelected.length);
        for (int i = origSelected.length; i < selected.length; i++) { 
            selected[i] = 0;
        }
        saveSelected();
    }

    public void clearData() { 
        byte[] sel = selected;
        for (int i = 0, n = sel.length; i < n; i++) { 
            sel[i] = 0;
        }
        saveSelected();
    }

    public void setData() { 
        byte[] sel = selected;
        for (int i = 0, n = sel.length; i < n; i++) { 
            sel[i] = (byte)-1;
        }
        saveSelected();
    }

    protected void destroyApp(boolean unconditional) {
        if (store != null) { 
            try { 
                store.closeRecordStore();
            } catch (Exception x) {
                criticalException(x);
            }
        }
        actionList = null;
    }    

    protected  void pauseApp() {
    }

    protected void startApp() {
        if (store == null) { 
            openStore();
        }
        Display.getDisplay(this).setCurrent(actionList);
    }

    public static void goBack() {
        instance.startApp();
    }

    public void commandAction(Command c, Displayable d) {
        if (c == QUIT_CMD) { 
            destroyApp(true);
            notifyDestroyed();
        } else { 
            int pos = actionList.getSelectedIndex();
            try {
                Displayable current = (Displayable)actions[pos].newInstance();
                Display.getDisplay(this).setCurrent(current);
            }
            catch(Exception err) {
                criticalException(err);
            }
        }
    }

    void criticalException(Exception x) {         
        x.printStackTrace();
        criticalError(x.toString());
    }
    
    void criticalError(String msg) {
        System.err.println(msg);
        error(msg);
    }

    boolean isSelected(Product product) { 
        int i = nPredefined + product.recordId;
        return (selected[i>>3] & (1 << (i & 7))) != 0;
    }

    boolean setSelected(Product product, boolean select) { 
        int i = nPredefined + product.recordId;
        if (((selected[i>>3] & (1 << (i & 7))) != 0) ^ select) { 
            selected[i>>3] ^= (1 << (i & 7));
            return true;
        }
        return false;
    }

    void saveSelected() { 
         try { 
            store.setRecord(SELECTED_ID, selected, 0, selected.length);            
        } catch (Exception x) { 
            criticalException(x);
            return;
        }
    }

    void error(String msg) {
        Alert a = new Alert(Locale.current.error);
        a.setType(AlertType.ERROR);
        a.setString(msg);
        a.setTimeout(Alert.FOREVER);
        Display.getDisplay(this).setCurrent(a);
    }
}
